;;;; tinyclos.scm

(cond-expand (hygienic-macros

; From '@' by Dan Muresan
(define-syntax slot@
  (syntax-rules (=)
    ((_ o) o)
    ((_ o slot = v) (slot-set! o 'slot v))
    ((_ o slot . slots) (slot@ (slot-ref o 'slot) . slots))))

(define-syntax define-class
  (syntax-rules ()
    [(_ name () slots)
     (define-class name (<object>) slots) ]
    [(_ name supers slots)
     (define-class name supers slots <class>) ]
    [(_ name () slots meta)
     (define-class name (<object>) slots meta) ]
    [(_ cname (supers ...) (slots ...) meta)
     (define cname (make meta 'name 'cname 'direct-supers (list supers ...) 'direct-slots (list 'slots ...))) ] ) )

(define-syntax define-generic
  (syntax-rules ()
    [(_ n class) (define n (make class 'name 'n))]
    [(_ n) (define n (make-generic 'n))] ) )

(define-macro (define-method head . body)
  (##sys#check-syntax 'define-method head '(symbol . _))
  (##sys#check-syntax 'define-method body '#(_ 1))
  (let gather ([args (##sys#slot head 1)]
               [specs '()]
               [vars '()] )
    (if (or (not (pair? args))
            (memq (car args) '(#!optional #!key #!rest)) )
        (let ([name (##sys#slot head 0)])
          `(add-method
            ,name
            (make-method
             (list ,@(reverse specs))
             (lambda (call-next-method ,@(reverse vars) ,@args) ,@body) ) ) )
        (let ([arg (##sys#slot args 0)])
          (gather (##sys#slot args 1)
                  (cons (if (pair? arg) (cadr arg) '<top>) specs)
                  (cons (if (pair? arg) (car arg) arg) vars) ) ) ) ) )

)(else

; From '@' by Dan Muresan
(define-macro (slot@ o . rest)
  (match rest
    [() o]
    [(slot '= v) `(slot-set! ,o ',slot ,v)]
    [(slot . slots) `(slot@ (slot-ref ,o ',slot) ,@slots)]) )

(define-macro (define-class name supers slots . meta)
  (##sys#check-syntax 'define-class name 'symbol)
  (##sys#check-syntax 'define-class supers '#(_ 0))
  (##sys#check-syntax 'define-class slots '#(_ 0))
  (##sys#check-syntax 'define-class meta '#(_ 0 1))
  `(##core#set! ,name
     (make ,(if (pair? meta) (##sys#slot meta 0) '<class>)
       'name ',name
       'direct-supers (list ,@(if (null? supers) '(<object>) supers))
       'direct-slots (list ,@(map (lambda (s) `',s) slots)) ) ) )

(define-macro (define-generic name . class)
  (let ((class (if (pair? class) (car class) '<generic>)))
    (##sys#check-syntax 'define-generic name 'symbol)
    `(define ,name (make ,class 'name ',name)) ) )

(define-macro (define-method head . body)
  (##sys#check-syntax 'define-method head '(symbol . _))
  (##sys#check-syntax 'define-method body '#(_ 1))
  (let gather ([args (##sys#slot head 1)]
               [specs '()]
               [vars '()] )
    (if (or (not (pair? args))
            (memq (car args) '(#!optional #!key #!rest)) )
        (let ([name (##sys#slot head 0)])
          `(##core#set! ,name
                        (##tinyclos#add-global-method
                         (##core#global-ref ,name)
                         ',name
                         (list ,@(reverse specs))
                         (##core#named-lambda ,name (call-next-method ,@(reverse vars) ,@args) ,@body) ) ) )
        (let ([arg (##sys#slot args 0)])
          (gather (##sys#slot args 1)
                  (cons (if (pair? arg) (cadr arg) '<top>) specs)
                  (cons (if (pair? arg) (car arg) arg) vars) ) ) ) ) )

))

;; For system use in extending the set of "builtin" classes.

(define-for-syntax (##tinyclos#make-classname-symbol str)
  (string->symbol (string-append "<" str ">")) )

(define-macro (define-primitive-class name pred . sclasses)
  `(define ,(##tinyclos#make-classname-symbol name)
     (##tinyclos#new-primitive-class ,name ,pred ,@sclasses)) )

(define-macro (define-structure-class name tag)
  `(define ,(##tinyclos#make-classname-symbol name)
     (##tinyclos#new-structure-class name ',tag)) )

(define-macro (define-tagged-pointer-class name tag)
  `(define ,(##tinyclos#make-classname-symbol name)
     (##tinyclos#new-tagged-pointer-class name ',tag)) )

(define-macro (define-extended-procedure-class name pred)
  `(define ,(##tinyclos#make-classname-symbol name)
     (##tinyclos#new-extended-procedure-class name pred)) )
